"""An implementation of the CloudEvent spec."""

import datetime
import typing
import uuid

import pendulum
import pydantic
from outcome.eventkit.data import CloudEventData

non_empty_string = pydantic.constr(strip_whitespace=True, min_length=1)


def generate_event_id() -> str:
    return str(uuid.uuid4())


class CloudEventV1_0(pydantic.BaseModel):  # noqa: N801, WPS214
    """This is an implementation of the v1.0 cloud event spec.

    https://github.com/cloudevents/spec/blob/v1.0/spec.md
    """

    class Config:
        title = 'CloudEvent@V1.0'
        validate_all = True
        allow_population_by_field_name = False
        validate_assignment = True
        extra = pydantic.Extra.forbid

    # The event's unique ID
    id: non_empty_string = pydantic.Field(default_factory=generate_event_id)  # noqa: WPS125, A003
    # The event's origin
    source: non_empty_string

    # Hardcoded to 1.0
    spec_version: typing.Literal['1.0'] = pydantic.Field('1.0', alias='specversion')

    # The event type, used to provide semantic meaning to the event
    type: non_empty_string  # noqa: WPS125, A003

    @property
    def data_content_type(self) -> typing.Optional[str]:
        if self.data:
            return self.data.data_content_type
        return None

    @property
    def data_schema(self) -> typing.Optional[str]:
        if self.data:
            return self.data.data_schema
        return None

    subject: typing.Optional[non_empty_string] = None
    time: typing.Optional[datetime.datetime] = pydantic.Field(default_factory=pendulum.now)

    @pydantic.validator('time')
    def validate_time(cls, value) -> typing.Optional[datetime.datetime]:  # noqa: N805
        if value is None:
            return None
        return pendulum.instance(value)

    # Data in its decoded form, i.e. some object
    data: typing.Optional[CloudEventData] = None

    @pydantic.validator('data')
    def validate_data(cls, value, values) -> CloudEventData:  # noqa: N805
        if value is not None:
            value.validate_data()
        return value

    @property
    def attributes(self) -> typing.Dict[str, typing.Any]:
        non_attribute_keys = {'data'}
        attributes = {
            attr: value for attr, value in self.dict(by_alias=True, exclude_none=True).items() if attr not in non_attribute_keys
        }

        if self.data:
            if self.data.data_content_type is not None:
                attributes['datacontenttype'] = self.data.data_content_type
            if self.data.data_schema is not None:
                attributes['dataschema'] = self.data.data_schema

        return attributes


CloudEvent = CloudEventV1_0



"""Basic tools to register event handlers and dispatch events."""

from collections import defaultdict
from typing import Callable, Dict, List

from outcome.eventkit.event import CloudEvent

CloudEventHandler = Callable[[CloudEvent], None]
CloudEventHandlerRegistry = Dict[str, List[CloudEventHandler]]

cloud_event_handler_registry: CloudEventHandlerRegistry = defaultdict(list)


def register_handler(
    event_type: str, handler: CloudEventHandler, registry: CloudEventHandlerRegistry = cloud_event_handler_registry,
) -> None:
    """Register a handler for an event type.

    Args:
        event_type (str): The event type
        handler (CloudEventHandler): The event handler.
        registry (CloudEventHandlerRegistry): The handler registry. Defaults to the global registry.
    """
    registry[event_type].append(handler)


def handles_events(*event_types: str, registry: CloudEventHandlerRegistry = cloud_event_handler_registry):
    """A decorator that registers the decorated function as a handler for the specified event types.

    Args:
        event_types (*str): The event types.
        registry (CloudEventHandlerRegistry): The handler registry. Defaults to the global registry.

    Returns:
        Callable[[CloudEventHandler], CloudEventHandler]: The decorator.
    """

    def handles_events_decorator(fn: CloudEventHandler) -> CloudEventHandler:
        for event_type in event_types:
            register_handler(event_type, fn, registry)

        return fn

    return handles_events_decorator


def dispatch(event: CloudEvent, registry: CloudEventHandlerRegistry = cloud_event_handler_registry) -> None:
    """Send an event to all of the registered handlers.

    Args:
        event (CloudEvent): The event to dispatch.
        registry (CloudEventHandlerRegistry): The handler registry. Defaults to the global registry.
    """
    for handler in cloud_event_handler_registry[event.type]:
        handler(event)



"""A class to represent data that can be encoded."""

from typing import Any, Optional

import pydantic
from outcome.eventkit.data.coder import DataCoder, DecodedData, EncodedData
from outcome.eventkit.mime import parse_mime_type


class CloudEventData(pydantic.BaseModel):
    # What is the data payload's format?
    # This is a MIME type that can be used to retrieve the decoder/encoder/validator functions
    # It's important to note that the event's payload can be in a completely different format
    # to the rest of the event data. For example, the event can be sent in JSON format, with
    # the data payload being encoded as a B64 XML string inside the JSON

    class Config:
        validate_all = True
        validate_assignment = True
        extra = pydantic.Extra.forbid

    data_content_type: Optional[str] = None
    data_schema: Optional[str] = None

    # The data itself, in a decoded form
    data: DecodedData = None

    def __eq__(self, o: Any) -> bool:
        if not isinstance(o, CloudEventData):  # pragma: no cover
            return False

        return o.data == self.data and o.data_content_type == self.data_content_type and o.data_schema == self.data_schema

    @pydantic.validator('data_content_type')
    def validate_data_content_type(cls, value) -> str:  # noqa: N805
        if value is None:
            return None
        return parse_mime_type(value).name

    @property
    def encoded_data(self) -> EncodedData:
        if self.data is None:
            return None

        if self.data_content_type is None:
            raise ValueError('Cannot encode data without content type')

        coder = self.get_coder(self.data_content_type)

        return coder.encode(self.data, self.data_content_type)

    @classmethod
    def from_encoded(
        cls, encoded_data: EncodedData, data_content_type: str, data_schema: Optional[str] = None,
    ) -> 'CloudEventData':
        data_content_type = parse_mime_type(data_content_type).name
        coder = cls.get_coder(data_content_type)
        decoded_data = coder.decode(encoded_data, data_content_type)

        if data_schema:
            coder.validate(decoded_data, data_content_type, data_schema)

        return cls(data=decoded_data, data_content_type=data_content_type, data_schema=data_schema)

    @staticmethod
    def get_coder(data_content_type: str) -> DataCoder:  # noqa: WPS602
        try:
            return DataCoder.data_content_types[data_content_type]
        except KeyError:
            raise ValueError(f'Unknown content type {data_content_type}')

    def validate_data(self) -> None:
        if self.data_content_type:
            coder = self.get_coder(self.data_content_type)
            coder.validate(self.data, self.data_content_type, self.data_schema)



"""Tools to build HTTP messages from CloudEvents."""

import re
import urllib
from typing import Any, Dict, List, Optional, Type, Union, cast

import pendulum
from outcome.eventkit.data import CloudEventData
from outcome.eventkit.event import CloudEvent
from outcome.eventkit.formats import CloudEventFormat
from outcome.eventkit.mime import parse_mime_type
from requests.structures import CaseInsensitiveDict
from requests.utils import check_header_validity

HeaderDict = Dict[str, str]


_header_attribute_prefix = 'ce-'
_header_attribute_prefix_pattern = r'^ce-'

_content_type_header = 'Content-Type'


class HTTPEvent:
    """A basic container for HTTP headers and a body."""

    headers: HeaderDict
    body: Optional[Union[bytes, str]]

    def __init__(self, body: Optional[Union[bytes, str]] = None, headers: Optional[HeaderDict] = None) -> None:
        self.headers: HeaderDict = CaseInsensitiveDict(headers or {})
        self.body = body


def attributes_to_headers(attributes: Dict[str, Any]) -> HeaderDict:
    """Constructs a header dict from the attributes of an event.

    `data` and `datacontenttype` are excluded.

    Args:
        attributes (Dict[str, Any]): The event attributes.

    Returns:
        HeaderDict: The header dict.
    """
    exclude_from_headers = {'datacontenttype'}

    headers = CaseInsensitiveDict()

    for attr, value in attributes.items():
        if attr not in exclude_from_headers:
            # Time is the only attribute in the spec that isn't represented
            # by a python str-like object
            if attr == 'time':
                value = pendulum.instance(value).isoformat()

            header = f'{_header_attribute_prefix}{attr.lower()}'

            check_header_validity((header, value))

            headers[header] = value

    return headers


# Method to attempt to parse a cloud event from either a binary or structured HTTP message
def from_http(http_event: HTTPEvent) -> CloudEvent:
    if looks_like_binary(http_event):
        return BinaryHTTPBinding.from_http(http_event)

    return StructuredHTTPBinding.from_http(http_event)


def looks_like_binary(http_event: HTTPEvent) -> bool:
    # Not an exact science, but we can check for some headers
    expected_headers = {'Content-Type', 'ce-specversion', 'ce-id', 'ce-type', 'ce-source'}

    # We can't use set intersection, since http_event.headers does additional case-insensitive checking
    if not all(header in http_event.headers for header in expected_headers):
        return False

    # The presence of the headers is not sufficient, since structured HTTP bindings can also contain the
    # headers, instead we must check the Content-Type header to ensure it's not of the type application/cloudevents
    # which is mandated by the spec
    content_type = parse_mime_type(http_event.headers['Content-Type'])

    return not (content_type.type == 'application' and content_type.subtype == 'cloudevents')


class BinaryHTTPBinding:
    """Creates a HTTP message in binary format.

    - Event `datacontenttype` is mapped to `Content-Type`
    - Other event attributes are mapped to HTTP headers, prefixed with `ce-`
    - Event data is serialized and used as the HTTP body

    See more at https://github.com/cloudevents/spec/blob/v1.0/http-protocol-binding.md#31-binary-content-mode
    """

    @staticmethod
    def to_http(event: CloudEvent) -> HTTPEvent:  # noqa: WPS602
        if event.data is not None and event.data_content_type is None:
            raise ValueError('Cannot construct a binary HTTP message from an event without a data content type')

        headers = attributes_to_headers(event.attributes)
        body = None

        if event.data:
            headers[_content_type_header] = event.data_content_type
            body = event.data.encoded_data

        return HTTPEvent(body, headers)

    @staticmethod
    def from_http(http_event: HTTPEvent) -> CloudEvent:  # noqa: WPS602
        # We don't want to add data_schema as an attribute of the event itself
        # since its attached to the CloudEventData instance instead
        excluded_attributes = {'ce-dataschema'}

        # We want to ignore case, since we don't know what case we're dealing with
        attributes = {
            re.sub(_header_attribute_prefix_pattern, '', attr, flags=re.I): urllib.parse.unquote(value)
            for attr, value in http_event.headers.items()
            if attr.lower().startswith(_header_attribute_prefix) and attr.lower() not in excluded_attributes
        }

        data = None
        data_content_type = http_event.headers.get(_content_type_header)
        data_schema = http_event.headers.get('ce-dataschema')

        if http_event.body is not None:
            if data_content_type is None:
                raise ValueError('Cannot create event from binary HTTP message without a Content-Type header')

            coder = CloudEventData.get_coder(data_content_type)
            decoded_data = coder.decode(http_event.body, data_content_type)

            data = CloudEventData(data=decoded_data, data_content_type=data_content_type, data_schema=data_schema)

        if data is None and (data_content_type or data_schema):
            data = CloudEventData(data_content_type=data_content_type, data_schema=data_schema)

        return CloudEvent(**{'data': data, **attributes})


class StructuredHTTPBinding:
    """Creates a structured HTTP message.

    The entire event is transmitted in the HTTP body.
    The provided `event_format` determines how the event is serialized.
    """

    @staticmethod
    def to_http(  # noqa: WPS602
        event: CloudEvent, event_format: Type[CloudEventFormat], include_attributes_in_headers: bool = False,
    ) -> HTTPEvent:

        if event_format.format_content_type is None:  # pragma: no cover
            raise ValueError('Event format needs to specify a content type')

        if include_attributes_in_headers:
            headers = attributes_to_headers(event.attributes)
        else:
            headers = cast(HeaderDict, CaseInsensitiveDict())

        headers[_content_type_header] = event_format.format_content_type
        body = event_format.encode(event)

        return HTTPEvent(body, headers)

    @staticmethod
    def from_http(http_event: HTTPEvent) -> List[CloudEvent]:  # noqa: WPS602
        try:
            format_content_type = http_event.headers[_content_type_header]
        except KeyError:
            raise ValueError('The HTTP event does not contain a content-type')

        try:
            event_format = CloudEventFormat.format_content_types[format_content_type]
        except KeyError:
            raise ValueError(f'Unknown format content type {format_content_type}')

        return event_format.decode(http_event.body)



"""An abstract data encoder/decoder used to serialize the data payload."""

from typing import Any, ClassVar, Optional, Union

from outcome.eventkit.mime import MIMETypeDict

EncodedData = Optional[Union[str, bytes]]
DecodedData = Optional[Any]


class DataCoder:

    # This is a mapping of data payload content types to the class that can handle the data payload
    # Content types can be added to this dict
    data_content_types: ClassVar[MIMETypeDict['DataCoder']] = MIMETypeDict['DataCoder']()

    @classmethod
    def encode(cls, data: DecodedData, content_type: str) -> EncodedData:  # pragma: no cover
        raise NotImplementedError

    @classmethod
    def decode(cls, encoded_data: EncodedData, content_type: str) -> DecodedData:  # pragma: no cover
        raise NotImplementedError

    @classmethod
    def validate(cls, data: DecodedData, content_type: str, schema_name: Optional[str] = None) -> None:  # pragma: no cover
        raise NotImplementedError



"""Tools to parse and register MIME types."""

from typing import Dict, List, Optional, Tuple, TypeVar

from lark import Lark, Tree
from lark.visitors import Interpreter
from outcome.utils.transformer_dict import TransformerDict

type_separator = '/'
subtype_separator = '.'
suffix_separator = '+'
param_separator = ';'


class MIMEType:
    type: str  # noqa: WPS125, A003
    subtype: str
    suffix: Optional[str]
    parameters: Dict[str, str]

    def __init__(self, default_charset: Optional[str]) -> None:
        if default_charset:
            self.parameters = {'charset': default_charset.lower()}
        else:
            self.parameters = {}
        self.suffix = None

    @property
    def charset(self) -> Optional[str]:
        return self.parameters.get('charset', None)

    @property
    def name(self) -> str:
        name = type_separator.join([self.type, self.subtype])

        if self.suffix:
            name = f'{name}+{self.suffix}'

        if self.parameters:
            params = param_separator.join(sorted(f'{key}={val}'.lower() for key, val in self.parameters.items()))
            name = f'{name}{param_separator}{params}'

        return name  # noqa: R504

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, MIMEType):  # pragma: no cover
            return False
        return o.name == self.name

    def __repr__(self) -> str:  # pragma: no cover
        return self.name

    def __hash__(self) -> int:  # pragma: no cover
        return hash(self.name)


class MIMEParseTreeInterpreter(Interpreter):
    """This class interprets the AST that results from the MIME parser.

    As the nodes of the AST are encountered, the MIMEType instance is built.

    The `type`, `subtype`, etc. methods are called for each node with the
    same name in the AST.
    """

    def __init__(self, default_charset: Optional[str]):
        self.mime_type = MIMEType(default_charset)

    def type(self, tree: Tree) -> None:  # noqa: WPS125, A003
        # There is only one root type, and it only has one token
        self.mime_type.type = tree.children[0].value.lower()  # noqa: WPS125

    def subtype(self, tree: Tree) -> None:
        # A subtype is made up of a sequence of dot-separated terms, followed
        # by an optional suffix
        children: List[Tree] = []
        for child in tree.children:
            if child.data == 'subtype_suffix':
                # The suffix only has one token
                self.mime_type.suffix = child.children[0].value.lower()
            else:
                children.append(child)

        # Each part of the subtype only has one token
        self.mime_type.subtype = subtype_separator.join(c.children[0].value for c in children).lower()

    def parameter(self, tree: Tree) -> None:
        # The parameters are made up of two tokens, after stripping whitespace
        param_key_token, param_value_token = [t for t in tree.children if not (isinstance(t, Tree) and t.data == 'whitespace')]

        param_key = param_key_token.value.lower()
        param_value = param_value_token.value.lower()

        self.mime_type.parameters[param_key] = param_value


_parser = None
CacheKey = Tuple[str, Optional[str]]
_cache: Dict[CacheKey, MIMEType] = {}


def parse_mime_type(mime_type_str: str, default_charset: Optional[str] = 'utf-8') -> MIMEType:
    # It's easier to build a mini-parser than try to do this with regex
    global _parser
    if not _parser:
        _parser = Lark.open('mime.lark', rel_to=__file__)  # noqa: WPS442,WPS122

    key = (mime_type_str, default_charset)

    if key not in _cache:
        try:
            tree = _parser.parse(mime_type_str)  # noqa: WPS442,WPS122,WPS121
        except Exception:
            raise ValueError(f'Invalid MIME type: {mime_type_str}')

        interpreter = MIMEParseTreeInterpreter(default_charset)
        interpreter.visit(tree)

        _cache[key] = interpreter.mime_type

    return _cache.get(key)


T = TypeVar('T')


def normalize_key(key: str) -> str:
    return parse_mime_type(key).name


class MIMETypeDict(TransformerDict[str, T, str]):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, transformer=normalize_key, **kwargs)



"""An implementation of the CloudEvent JSON format."""

import base64
import json
from typing import Dict, Union

import pendulum
from outcome.eventkit.data import CloudEventData
from outcome.eventkit.event import CloudEvent
from outcome.eventkit.formats import CloudEventFormat
from outcome.eventkit.mime import parse_mime_type

content_type_name = parse_mime_type('application/cloudevents+json').name
json_content_type_name = parse_mime_type('application/json').name


class JSONCloudEventFormat(CloudEventFormat):
    @classmethod
    def encode(cls, event: CloudEvent) -> str:
        payload = event.attributes

        # According to the spec:
        #
        # If the event has data, but has no data content type, then it is assumed
        # that the content type is that of the envelope (in this case JSON).
        # https://github.com/cloudevents/spec/blob/v1.0/spec.md#datacontenttype (second to last paragraph)
        #
        # If the data content type is JSON, we don't encode the data, since it can
        # just be encoded with the rest of the envelope to form one JSON document.
        # https://github.com/cloudevents/spec/blob/v1.0/spec.md#type-system (last paragraph)
        if event.data:
            if event.data_content_type is None or event.data_content_type == json_content_type_name:
                payload['datacontenttype'] = json_content_type_name
                payload['data'] = event.data.data  # noqa: WPS204
            else:
                payload['data'] = event.data.encoded_data

        # If data is binary, we need to encode it and store it
        # in a separate field
        # https://github.com/cloudevents/spec/blob/v1.0/json-format.md#31-handling-of-data
        try:
            if isinstance(payload['data'], bytes):
                payload['data_base64'] = base64.b64encode(payload.pop('data')).decode('utf-8')

            # We handle key errors since data is optional
        except KeyError:
            pass

        try:
            # We use pendulum to ensure timezones are handled correctly
            payload['time'] = pendulum.instance(payload['time']).isoformat()
            # We handle key errors since timestamp is optional
        except KeyError:
            pass

        return json.dumps(payload)

    @classmethod
    def decode(cls, raw_event: Union[bytes, str]) -> CloudEvent:
        payload: Dict[str] = json.loads(raw_event)

        try:
            payload['data'] = base64.b64decode(payload.pop('data_base64')).decode('utf-8')  # noqa: WPS204
        except KeyError:
            pass

        # If we have data, we need to unpack it
        if payload.get('data') is not None:
            data_schema = payload.pop('dataschema', None)

            # if there was no content type, or it was JSON, it's already unpacked
            data_content_type = payload.pop('datacontenttype', json_content_type_name)
            if data_content_type == json_content_type_name:
                payload['data'] = CloudEventData(
                    data=payload['data'], data_content_type=json_content_type_name, data_schema=data_schema,
                )
            else:
                payload['data'] = CloudEventData.from_encoded(
                    payload['data'], data_content_type=data_content_type, data_schema=data_schema,
                )

        # Parse with pendulum to get the timezone right
        try:
            payload['time'] = pendulum.parse(payload['time'])
        except KeyError:
            pass

        return CloudEvent(**payload)


JSONCloudEventFormat.format_content_type = content_type_name
cloud_event_format = JSONCloudEventFormat



from outcome.eventkit.data import CloudEventData
from outcome.eventkit.data import json as json_data_coder
from outcome.eventkit.data.coder import DataCoder
from outcome.eventkit.event import CloudEvent
from outcome.eventkit.formats import CloudEventFormat
from outcome.eventkit.formats import json as json_format

# Register known data content types
DataCoder.data_content_types[json_data_coder.content_type_name] = json_data_coder.coder


# Register known cloud event format types
CloudEventFormat.format_content_types[json_format.content_type_name] = json_format.cloud_event_format

__all__ = ['CloudEventData', 'CloudEvent']



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['outcome',
 'outcome.eventkit',
 'outcome.eventkit.data',
 'outcome.eventkit.formats',
 'outcome.eventkit.mime',
 'outcome.eventkit.protocol_bindings']

package_data = \
{'': ['*']}

install_requires = \
['lark-parser>=0.10.1,<0.11.0',
 'outcome-utils>=4.16.1,<5.0.0',
 'pendulum>=2.1.2,<3.0.0',
 'pydantic>=1.7.2,<2.0.0',
 'requests>=2.24.0,<3.0.0']

setup_kwargs = {
    'name': 'outcome-eventkit',
    'version': '0.2.4',
    'description': 'A toolkit for emitting and handling events, following the CloudEvent spec.',
    'long_description': "# eventkit-py\n![ci-badge](https://github.com/outcome-co/eventkit-py/workflows/Release/badge.svg?branch=v0.2.4) ![version-badge](https://img.shields.io/badge/version-0.2.4-brightgreen)\n\nA toolkit for emitting and handling events, following the CloudEvent spec.\n\n## Installation\n\n```sh\npoetry add outcome-eventkit\n```\n\n## Usage\n\nIt's a good idea to be familiar with the [CloudEvent v1.0 spec](https://github.com/cloudevents/spec/blob/v1.0/spec.md) before using this library.\n\nThe library implements the following (but is extensible):\n\n- [The Core Event object](https://github.com/cloudevents/spec/blob/v1.0/spec.md)\n- [A JSON formatter](https://github.com/cloudevents/spec/blob/v1.0/json-format.md)\n- [The HTTP Protocol Bindings (Structured and Binary)](https://github.com/cloudevents/spec/blob/v1.0/http-protocol-binding.md)\n\n### Event Example\n```py\nfrom outcome.eventkit import CloudEvent, CloudEventData\nfrom outcome.eventkit.formats.json import JSONCloudEventFormat\nfrom outcome.eventkit.protocol_bindings.http import StructuredHTTPBinding, BinaryHTTPBinding\n\nimport requests\n\n# Create a payload, specifying how it will be encoded\ndata = CloudEventData(\n    data_content_type='application/json',\n    data={\n        'hello': 'world'\n    }\n)\n\n# Create an event\nevent = CloudEvent(type='co.outcome.events.sample', source='example', data=data)\n\n# Create a JSON representation of the event\nserialised_event = JSONCloudEventFormat.encode(event)\n\n# Or...\n\n# Create a HTTP message with the event\nhttp_message = BinaryHTTPBinding.to_http(event)\n\n# Or...\n\n# Create a structured HTTP message with the event and a JSON formatter\nhttp_message = StructuredHTTPBinding.to_http(event, JSONCloudEventFormat)\n\n# Post the event somewhere...\nrequests.post('http://example.org', headers=http_message.headers, data=http_message.body)\n```\n\n### Dispatch Example\n```py\nfrom outcome.eventkit import dispatch, CloudEvent\n\n# You can register event handlers that will be called\n# when a corresponding event is dispatched\n\n# Explicitly\ndef my_event_handler(event: CloudEvent) -> None:\n    ...\n\n# Register a handler for an event type\ndispatch.register_handler('co.outcome.event', my_event_handler)\n\n\n# Or via a decorator\n@dispatch.handles_events('co.outcome.event', 'co.outcome.other-event')\ndef my_other_handler(event: CloudEvent) -> None:\n    ...\n\n\n# Then notify all registered handlers of an event\nev = CloudEvent(type='co.outcome.event', source='example')\ndispatch.dispatch(ev)\n```\n\nBy default, all handlers are stored in a single registry, but you can provide your own registry to the `register_handler`/`handles_events`/`dispatch` methods with the `registry` keyword argument.\n\n```py\nfrom outcome.eventkit import dispatch\nfrom collections import defaultdict\n\nmy_registry: dispatch.CloudEventHandlerRegistry = defaultdict(list)\n\n@dispatch.handles_events('co.outcome.event', registry=my_registry)\ndef my_handler(event: CloudEvent) -> None:\n    ...\n\nev = CloudEvent(type='co.outcome.event', source='example')\ndispatch.dispatch(ev, registry=my_registry)\n```\n\n## Development\n\nRemember to run `./pre-commit.sh` when you clone the repository.\n",
    'author': 'Outcome Engineering',
    'author_email': 'engineering@outcome.co',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/outcome-co/eventkit-py',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8.6,<3.9.0',
}


setup(**setup_kwargs)



"""A bare-bones implementation of a JSON data payload coder."""

import json
from typing import Optional

from outcome.eventkit.data.coder import DataCoder, DecodedData, EncodedData


class JSONDataCoder(DataCoder):
    @classmethod
    def encode(cls, data: DecodedData, content_type: str) -> str:
        return json.dumps(data)

    @classmethod
    def decode(cls, encoded_data: EncodedData, content_type: str) -> DecodedData:
        return json.loads(encoded_data)

    @classmethod
    def validate(cls, data: DecodedData, content_type: str, schema_name: Optional[str]) -> None:  # pragma: no cover
        ...


content_type_name = 'application/json'
coder = JSONDataCoder



"""An abstract implementation of the CloudEvent format spec."""

from typing import ClassVar, Optional, Type, TypeVar, Union

from outcome.eventkit.event import CloudEvent
from outcome.eventkit.mime import MIMETypeDict, parse_mime_type

T = TypeVar('T')

CloudEventFormatMIMETypeDict = MIMETypeDict[Type['CloudEventFormat']]


class CloudEventFormatMeta(type):
    # Using a metaclass, we can implement class-level properties
    @property
    def format_content_type(cls) -> Optional[str]:  # noqa: N805
        try:
            return cls._format_content_type
        except AttributeError:
            return None

    @format_content_type.setter
    def format_content_type(cls, content_type: Optional[str]) -> None:  # noqa: N805
        if content_type:
            cls._format_content_type = parse_mime_type(content_type).name
        else:
            cls._format_content_type = None


class CloudEventFormat(metaclass=CloudEventFormatMeta):
    # The registry of all known format_content_types
    format_content_types: ClassVar[CloudEventFormatMIMETypeDict] = CloudEventFormatMIMETypeDict()

    @classmethod
    def encode(cls, event: CloudEvent) -> Union[bytes, str]:  # pragma: no cover
        raise NotImplementedError

    @classmethod
    def decode(cls, raw_event: Union[bytes, str]) -> CloudEvent:  # pragma: no cover
        raise NotImplementedError
